#include "InputMap.hpp"

const robin_hood::unordered_map<Input, int> DEFAULT_MAP = {
    {Input::Forward, GLFW_KEY_W},
    {Input::Backward, GLFW_KEY_S},
    {Input::Left, GLFW_KEY_A},
    {Input::Right, GLFW_KEY_D},
    {Input::Up, GLFW_KEY_SPACE},
    {Input::Down, GLFW_KEY_LEFT_SHIFT},
    {Input::Mouse, GLFW_KEY_E},
    {Input::Debug, GLFW_KEY_F3},
    {Input::Throw, GLFW_KEY_B},
    {Input::Quit, GLFW_KEY_ESCAPE}
};

InputMap::InputMap(GLFWwindow* window, const std::vector<std::pair<int, int>>& keys): window(window), map(DEFAULT_MAP) {
    for(const auto& key: keys) {
        map.insert_or_assign(static_cast<Input>(key.first), key.second);
    }
}
InputMap::~InputMap() { }

void InputMap::exportKeys(std::vector<std::pair<int, int>>& keys) const {
    keys.clear();
    for(const auto& key: map) {
        if (const auto it = DEFAULT_MAP.find(key.first); it == DEFAULT_MAP.end() || it->second != key.second) {
            keys.emplace_back((int)key.first, key.second);
        }
    }
}

void InputMap::saveKeys() {
    for (auto input : toggles) {
        previous.insert_or_assign(input, isDown(input));
    }
    previousLeft = isDown(Mouse::Left);
    previousRight = isDown(Mouse::Right);
}

bool InputMap::isDown(Input input) const {
    return glfwGetKey(window, map.at(input)) == GLFW_PRESS;
}
bool InputMap::wasDown(Input input) const {
    return previous.at(input);
}
bool InputMap::isPressing(Input input) const {
    return isDown(input) && !wasDown(input);
}
bool InputMap::isReleasing(Input input) const {
    return !isDown(input) && wasDown(input);
}

void InputMap::toggle(bool &val, Input input) const {
    if(isPressing(input))
        val = !val;
}

bool InputMap::isDown(Mouse input) const {
    return glfwGetMouseButton(window, static_cast<int>(input)) == GLFW_PRESS;
}
bool InputMap::wasDown(Mouse input) const {
    switch (input) {
    case Mouse::Left:
        return previousLeft;

    case Mouse::Right:
        return previousRight;

    default:
        return false;
    }
}
bool InputMap::isPressing(Mouse input) const {
    return isDown(input) && !wasDown(input);
}
bool InputMap::isReleasing(Mouse input) const {
    return !isDown(input) && wasDown(input);
}

void InputMap::poll() const { glfwPollEvents(); }