#pragma once

#include <vector>
#include <algorithm>
#include <robin_hood.h>

namespace data {

/// Queue always sorted by incressing priority (pop: max priority)
/// K: key, W: weight, Unique: remove duplicates
/// Size: Optional size bound, Reverse: descressing priority (pop: min)
///NOTE: Not thread safe
template <class K, class W, bool Unique = false, std::size_t Size = 0, bool Reverse = false>
class sorted_queue {
private:
    std::vector<std::pair<K, W>> vec;

    static bool cmpByWeight(const W &a, const W &b) {
        return (a < b) != Reverse;
    }

public:
    void push(const K& key, const W& weight) {
        if constexpr (Unique) {
            // Remove previous
            for (auto it = vec.begin(); it != vec.end(); ++it) {
                if (it->first == key) {
                    vec.erase(it);
                    break;
                }
            }
        }

        if constexpr (Size > 0) {
            // Keep under Size
            if (size() >= Size) {
                if (cmpByWeight(weight, vec.front().second))
                    return;
                vec.erase(vec.begin());
            }
        }

        auto it = vec.begin();
        while (it != vec.end() && !cmpByWeight(weight, it->second)) {
            ++it;
        }
        vec.insert(it, std::make_pair(key, weight));
    }

    bool pop(K& out) {
        if (empty())
            return false;

        out = vec.back().first;
        vec.pop_back();
        return true;
    }

    bool empty() const {
        return vec.empty();
    }

    size_t size() const {
        return vec.size();
    }

};
}