#pragma once

#include "Area.hpp"
#include "../generational/map.hpp"
#include "../generational/hierarchy.hpp"

namespace world {

    /// ECS data (aka all universe readable knowledge)
    class Elements {
    public:
        /// All element in universe
        generational::vector<world::Node> nodes;
        /// Relations (orbits) of nodes
        generational::hierarchy<world::Node, world::Hierarchy> hierarchy;
        //MAYBE: split this big struct in internal SoA
        using hierarchy_entry = generational::hierarchy<Node, Hierarchy>::entry;
        generational::map<world::Node> areas;
        generational::map<world::Node> parts;
        generational::vector<world::Model> models;

        // MAYBE: move to Node
        static constexpr size_t TYPE_SIZE = 2;
        static constexpr uint8_t TYPE_MASK = ((1 << TYPE_SIZE)-1);
        enum class Type: uint8_t {
            None = 0,
            Instance = 1,
            Part = 2,
            Area = 3,
        };
        static constexpr Type GetType(const node_id &i) { return static_cast<Type>(i.val.flag() & TYPE_MASK); }
        template <Type T>
        static constexpr bool Is(const node_id &i) { return GetType(i) == T; }

        enum class Flag: uint8_t {
            Moved = 0, /// Nodes with modified state (transform, velocity or hierarchy)
            Removed = 1, /// Freed nodes
            Added = 2, /// New node or modified content
            Fixed = 3, /// Without hierarchy
            ChunkLoader = 4, /// Load chunks around
            MAX = 8-TYPE_SIZE-1
        };
        template <Flag F>
        static constexpr bool Has(const node_id &i) { return (i.val.flag() >> (TYPE_SIZE + static_cast<int>(F))) & 1; }
        template <Flag F, bool V = true>
        static constexpr uint8_t Set(uint8_t flag) {
            constexpr uint8_t IDX = TYPE_SIZE + static_cast<int>(F);
            return (flag & ~static_cast<uint8_t>(1 << IDX)) | (V << IDX);
        }
        template <Flag F, bool V = true>
        static constexpr node_id Set(const node_id &i) {
            return generational::id(i.val.index(), i.val.generation(), Set<F, V>(i.val.flag()));
        }

        node_id withFlag(node_ref r) const { return nodes.with_flag(r); }
        Node *findNode(node_id i) { return nodes.directly_at(i); }
        const Node* findNode(node_id i) const { return nodes.directly_at(i); }
        Node* findNode(node_ref r) { return nodes.directly_at(r); }
        NodeOf<Area>* findArea(area_id i) {
            assert(Is<Type::Area>(i.val));
            return NodeOf<Area>::Make(findNode(i.val));
        }
        const NodeOf<Area>* findArea(area_id i) const {
            assert(Is<Type::Area>(i.val));
            return NodeOf<Area>::Make(findNode(i.val));
        }
        NodeOf<Part>* findPart(part_id i) {
            assert(Is<Type::Part>(i.val));
            return NodeOf<Part>::Make(findNode(i.val));
        }
        const NodeOf<Part>* findPart(part_id i) const {
            assert(Is<Type::Part>(i.val));
            return NodeOf<Part>::Make(findNode(i.val));
        }

        world_pos computeAbsolute(const node_ref &pref, const voxel_pos& rel) const;
        transform computeAbsolute(const node_id &pid, const transform& rel) const;
        /// Maybe null
        node_id findParent(const node_id &id) const;
    };

    /// Edittable Elements container
    class EdittableElements: public Elements {
    public:
        EdittableElements(bool loadModels = true);

        struct start_point {
            start_point(node_id p, const transform &r, const velocity &vel = FIXED):
                parent(p), relative(r), vel(vel) { }
            start_point(const relative_pos& p, const velocity &vel = FIXED):
                start_point(p.parent, transform(p.position), vel) { }

            node_id parent;
            transform relative;
            velocity vel;
        };

        /// Children are link to parent
        void remove(node_id);
        /// Also delete childrens
        void removeRecursive(node_id);
        /// Set parent and positions
        void move(node_id id, const relative_pos &p);
        /// Set parent, position and velocity
        void move(node_id id, const start_point &p);

    protected:
        void extract(node_ref);
        void insert(node_ref, const start_point& sp);

        node_id create(const start_point&, uint8_t flag, const std::shared_ptr<world::Element>&);
        node_id createAt(node_id, const start_point&, const std::shared_ptr<world::Element>&);
        node_id createAt(generational::id::idx_t, const start_point&, uint8_t flag, const std::shared_ptr<world::Element>&);
    };
}