#pragma once

#include "actions.hpp"
#include <queue>

namespace world::iterator {

using pair = std::pair<glm::ivec3, float>;

class AbstractFill {
public:
    virtual bool next(pair&) = 0;

protected:
    static constexpr uint32_t Diam(uint16_t radius) { return radius * 2 + 1; }
};
class AbstractBorder {
public:
    virtual bool next(glm::ivec3&) = 0;
};

/// From -radius to radius
std::unique_ptr<AbstractFill> Get(action::Shape, uint16_t radius);
/// Border sampling from -radius to radius
std::unique_ptr<AbstractBorder> GetBorder(action::Shape, uint16_t radius);

template <typename Chunk>
struct last_chunk {
    typename std::shared_ptr<Chunk> ptr = nullptr;
    chunk_pos pos;
};

template <typename Chunk, typename CC>
inline bool GetVoxel(CC &chunks, const std::pair<chunk_pos, chunk_voxel_idx> &split, Voxel &out, last_chunk<Chunk>& ck) {
    if (split.first != ck.pos) {
        if (chunks.inRange(split.first)) {
            if(const auto ptr = chunks.findInRange(split.first)) {
                ck.ptr = std::static_pointer_cast<Chunk>(ptr);
                ck.pos = split.first;
            } //TODO: handle not loaded chunks
        } else {
            ck.ptr = nullptr;
        }
    }
    if (ck.ptr != nullptr) {
        out = ck.ptr->get(split.second);
        return true;
    }
    return false;
}

constexpr auto FILL_DELAY = .05f;
/// Apply shape on single chunk container
/// TODO: multi-part
template<typename Chunk, typename CC, typename CB>
inline void Apply(CC &chunks, action::FillShape fill, const CB& callback) {
    const auto iterator = world::iterator::Get(fill.shape, fill.radius); //TODO: rotation
    last_chunk<Chunk> ck;
    Voxel prev;
    for (pair point; iterator->next(point);) {
        const voxel_pos offset = point.first;
        const auto split = glm::splitIdx(fill.pos.voxel + offset);
        if (GetVoxel<Chunk>(chunks, split, prev, ck)) {
            const auto next = prev.filled(fill.val, point.second);
            if (prev.value != next.value) {
                callback(ck.ptr, ck.pos, split.second, prev, next, glm::length2(offset) / fill.radius * FILL_DELAY);
            }
        }
    }
}
/// Find parts using flood fill
template<typename Chunk, typename CC, typename CB>
inline void Split(CC &chunks, action::FillShape fill, size_t floodFillLimit, const CB& callback) {
    if (floodFillLimit == 0 || fill.val.is_solid())
        return;

    last_chunk<Chunk> ck;
    Voxel cur;
    // Check for subpart break
    robin_hood::unordered_set<voxel_pos> joints;
    {
        const auto iterator = GetBorder(fill.shape, fill.radius+1);
        for (glm::ivec3 point; iterator->next(point);) {
            const auto full = fill.pos.voxel + voxel_pos(point);
            if (GetVoxel<Chunk>(chunks, glm::splitIdx(full), cur, ck) && cur.is_solid())
                joints.insert(full);
        }
    }
    constexpr std::array<voxel_pos, 6> DIRS = { //MAYBE: diags
        voxel_pos(1, 0, 0), voxel_pos(-1, 0, 0), voxel_pos(0, 1, 0),
        voxel_pos(0, -1, 0), voxel_pos(0, 0, 1), voxel_pos(0, 0, -1)};
    while (!joints.empty()) {
        std::queue<voxel_pos> todo;
        todo.push(*joints.begin());
        robin_hood::unordered_set<voxel_pos> part;
        part.insert(todo.front());
        for (size_t i = 0; !todo.empty() && i < floodFillLimit; i++) {
            const auto full = todo.front();
            todo.pop();
            joints.erase(full);
            if (GetVoxel<Chunk>(chunks, glm::splitIdx(full), cur, ck) && cur.is_solid()) {
                for (const auto& dir: DIRS) {
                    const auto off = full + dir;
                    if (part.insert(off).second)
                        todo.push(off);
                }
            }
        }
        if (todo.empty()) {
            callback(part);
        }
    }
}

/// Call Apply then clean Splitted
template<typename Chunk, typename CC, typename CB>
inline void ApplyEraseSplitted(CC &chunks, action::FillShape fill, size_t floodFillLimit, const Voxel& cleanVoxel, const CB& callback) {
    Apply<Chunk>(chunks, fill, callback);
    Split<Chunk>(chunks, fill, floodFillLimit, [&](const robin_hood::unordered_set<voxel_pos>& part) {
        last_chunk<Chunk> ck; Voxel prev;
        for(const auto& full: part) {
            const auto split = glm::splitIdx(full);
            if (world::iterator::GetVoxel<Chunk>(chunks, split, prev, ck) && prev.is_solid()) {
                if (prev.value != cleanVoxel.value) {
                    callback(ck.ptr, ck.pos, split.second, prev, cleanVoxel, fill.radius * FILL_DELAY);
                }
            }
        }
    });
}

class Cube final: public AbstractFill {
public:
    bool next(pair&) override;

    Cube(uint16_t radius): radius(radius), pos(-radius, -radius, -radius) { }

private:
    const uint16_t radius;
    glm::ivec3 pos;
};
class CubeBorder final: public AbstractBorder {
public:
    bool next(glm::ivec3&) override;

    CubeBorder(uint16_t radius): radius(radius), pos(-radius, -radius, -radius), face(0) { }

private:
    const uint16_t radius;
    glm::ivec3 pos;
    int face;
};

/// Interger sphere
class RawSphere final: public AbstractFill {
public:
    bool next(pair&) override;

    RawSphere(uint16_t radius): radius(radius), pos(-radius-1, 0, 0), dy(0), dz(0) { }

private:
    const uint16_t radius;
    glm::ivec3 pos;
    int dy;
    int dz;
};
class SphereBorder final: public AbstractBorder {
public:
    bool next(glm::ivec3&) override;

    SphereBorder(uint16_t radius): radius(radius), max_i(radius*3), max_j(radius*4), i(0), j(0) { }

private:
    const uint16_t radius;
    const uint32_t max_i;
    const uint32_t max_j;
    uint32_t i;
    uint32_t j;
};

/// Anti-aliased sphere
class SmoothSphere final: public AbstractFill {
public:
    bool next(pair&) override;

    SmoothSphere(uint16_t radius): radius(radius), pos(-radius-1, 0, 0), dy(0), dz(0) { }

private:
    const uint16_t radius;
    glm::ivec3 pos;
    double dy;
    double dz;
};
}